// Original by Sensebloom. Some modifications made to enable Scratch to recieve commands.
// Removed OSC code, added the Winsock UDP code.
// Stephen Howell (stephen.r.howell@gmail.com)
//
// Modified from source downloaded at http://stephen-howell.tumblr.com/
// by Mauro Brunato for communication with a UDP client
// meant to control a Lego Mindstorms NXT Tribot.

#include <winsock2.h>
#include <XnCppWrapper.h>

char *ADDRESS = "127.0.0.1";
int PORT = 42001;	// Scratch Specific port
WSADATA wsaData;
SOCKET SendSocket;
sockaddr_in RecvAddr;

const int BufLen = 2048;
char SendBuf[BufLen];

char tmp[50];
float jointCoords[3];

//Multipliers for coordinate system. Scratch values here
/*
double mult_x = 480;
double mult_y = 360;
double mult_z = 1;
*/
double mult_x = 2;
double mult_y = -2;
double mult_z = 2;

//Offsets for coordinate system. Scratch values here
/*
double off_x = -240;
double off_y = -180;
double off_z = 0.0;
*/
double off_x = -1;
double off_y = 1;
double off_z = -4;

bool mirrorMode = true;
int nDimensions = 3;

xn::UserGenerator userGenerator;
XnChar g_strPose[20] = "";

XnSkeletonJoint jointList[] = {
	XN_SKEL_HEAD, XN_SKEL_NECK, XN_SKEL_LEFT_COLLAR, XN_SKEL_LEFT_SHOULDER, XN_SKEL_LEFT_ELBOW, XN_SKEL_LEFT_WRIST,
	XN_SKEL_LEFT_HAND, XN_SKEL_LEFT_FINGERTIP, XN_SKEL_RIGHT_COLLAR, XN_SKEL_RIGHT_SHOULDER, XN_SKEL_RIGHT_ELBOW,
	XN_SKEL_RIGHT_WRIST, XN_SKEL_RIGHT_HAND, XN_SKEL_RIGHT_FINGERTIP, XN_SKEL_TORSO, XN_SKEL_WAIST, XN_SKEL_LEFT_HIP,
	XN_SKEL_LEFT_KNEE, XN_SKEL_LEFT_ANKLE, XN_SKEL_LEFT_FOOT, XN_SKEL_RIGHT_HIP, XN_SKEL_RIGHT_KNEE, XN_SKEL_RIGHT_ANKLE,
	XN_SKEL_RIGHT_FOOT };

char* jointNames[] = { "head", "neck", "l_collar", "l_shoulder", "l_elbow", "l_wrist", "l_hand", "l_fingertip", "r_collar",
	"r_shoulder", "r_elbow", "r_wrist", "r_hand", "r_fingertip", "torso", "waist", "l_hip", "l_knee", "l_ankle", "l_foot",
	"r_hip", "r_knee", "r_ankle", "r_foot" };


void broadcastScratch(char* message)
{
	//---------------------------------------------
	// Before we send each message, we need to send some size code
	unsigned char sizeBytes[4];
	int len = strlen(message);
	// Can't find the example page I found this on - think it was 'Catenary' who posted it(spelling?)
	sizeBytes[0] =(unsigned char)( len >> 24 );
	sizeBytes[1] =(unsigned char)( (len << 8) >> 24 );
	sizeBytes[2] =(unsigned char)( (len << 16) >> 24 );
	sizeBytes[3] =(unsigned char)( (len << 24) >> 24 );

	sendto(SendSocket, (char*)sizeBytes, 4, 0, (SOCKADDR *) &RecvAddr, sizeof(RecvAddr));
	sendto(SendSocket, message, strlen(message), 0, (SOCKADDR *) &RecvAddr, sizeof(RecvAddr));
}

// Callback: New user was detected
void XN_CALLBACK_TYPE User_NewUser(xn::UserGenerator& generator, XnUserID nId, void* pCookie) 
{
	userGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
	broadcastScratch("broadcast new_user");
}

// Callback: An existing user was lost
void XN_CALLBACK_TYPE User_LostUser(xn::UserGenerator& generator, XnUserID nId, void* pCookie) 
{
	userGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
	broadcastScratch("broadcast lost_user");
}

// Callback: Detected a pose
// I don't think this is needed by Scratch
void XN_CALLBACK_TYPE UserPose_PoseDetected(xn::PoseDetectionCapability& capability, const XnChar* strPose, XnUserID nId, void* pCookie) 
{
	printf("Pose %s detected for user %d\n", strPose, nId);
	userGenerator.GetPoseDetectionCap().StopPoseDetection(nId);
	userGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
}

// Callback: Started calibration
void XN_CALLBACK_TYPE UserCalibration_CalibrationStart(xn::SkeletonCapability& capability, XnUserID nId, void* pCookie) {
	printf("Calibration started for user %d\n", nId);
}

// Callback: Finished calibration
void XN_CALLBACK_TYPE UserCalibration_CalibrationEnd(xn::SkeletonCapability& capability, XnUserID nId, XnBool bSuccess, void* pCookie) {
	if (bSuccess) {
		printf("Calibration complete, start tracking user %d\n", nId);
		userGenerator.GetSkeletonCap().StartTracking(nId);
		broadcastScratch("broadcast tracking");
	}
	else 
	{
		printf("Calibration failed for user %d\n", nId);
		userGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
	}
}

int jointPos(XnUserID player, XnSkeletonJoint eJoint) 
{
	XnSkeletonJointPosition joint;
	userGenerator.GetSkeletonCap().GetSkeletonJointPosition(player, eJoint, joint);

	if (joint.fConfidence < 0.5)
		return -1;
	// The -1 is for Scratch
	jointCoords[0] = player;
	jointCoords[1] = off_x + (mult_x * (1280 - joint.position.X) / 2560);
	jointCoords[2] = off_y + (mult_y * (1280 - joint.position.Y) / 2560);
	jointCoords[3] = off_z + (mult_z * joint.position.Z * 7.8125 / 10000);
	return 0;
}

void send(const xn::DepthMetaData& dmd)
{
	XnUserID aUsers[15];
	XnUInt16 nUsers = 15;
	userGenerator.GetUsers(aUsers, nUsers);
	for (int i = 0; i < nUsers; ++i)
	{
		if (userGenerator.GetSkeletonCap().IsTracking(aUsers[i])) 
		{
			*SendBuf = '\0';
			for(int j = 0; j < 24 ; j++)
				if(jointPos(aUsers[i], jointList[j]) == 0)	
					sprintf(SendBuf + strlen(SendBuf), "%d\t%f\t%f\t%f\n", j, jointCoords[1], jointCoords[2], jointCoords[3]);
			sendto(SendSocket, SendBuf, strlen(SendBuf)+1, 0, (SOCKADDR *) &RecvAddr, sizeof(RecvAddr));
			puts (SendBuf);
		}
	}
}


void checkRetVal(XnStatus nRetVal) {
	if (nRetVal != XN_STATUS_OK) {
		printf("There was a problem initializing kinect... Make sure you have\
			   connected both usb and power cables and that the driver and OpenNI framework\
			   are correctly installed.\n\n");
		exit(1);
	}
}

int main(int argc, char **argv)
{
	printf("OpenNI (Kinect) to Scratch version 0.1\n");
	printf("Please see http://stephen-howell.tumblr.com for help using this program\n");
 
	printf("Original OSCeleton code by Sensebloom (info@sensebloom.com)\nScratch modifications by Stephen Howell (stephen.r.howell@gmail.com)");
	printf("\n\n\nAttempting to connect to the Kinect and send tracking information to Scratch...\n");

	xn::Context context;
	xn::DepthGenerator depth;
	context.Init();
	depth.Create(context);

	XnMapOutputMode mapMode;
	mapMode.nXRes = XN_VGA_X_RES;
	mapMode.nYRes = XN_VGA_Y_RES;
	mapMode.nFPS = 30;
	depth.SetMapOutputMode(mapMode);

	XnStatus nRetVal = XN_STATUS_OK;
	nRetVal = context.FindExistingNode(XN_NODE_TYPE_USER, userGenerator); // This crashes on my setupcif kinect not plugged in
	if (nRetVal != XN_STATUS_OK)
		nRetVal = userGenerator.Create(context);

	XnCallbackHandle hUserCallbacks, hCalibrationCallbacks, hPoseCallbacks;
	checkRetVal(userGenerator.RegisterUserCallbacks(User_NewUser, User_LostUser, NULL, hUserCallbacks));
	checkRetVal(userGenerator.GetSkeletonCap().RegisterCalibrationCallbacks(UserCalibration_CalibrationStart, UserCalibration_CalibrationEnd, NULL, hCalibrationCallbacks));
	checkRetVal(userGenerator.GetPoseDetectionCap().RegisterToPoseCallbacks(UserPose_PoseDetected, NULL, NULL, hPoseCallbacks));
	checkRetVal(userGenerator.GetSkeletonCap().GetCalibrationPose(g_strPose));
	checkRetVal(userGenerator.GetSkeletonCap().SetSkeletonProfile(XN_SKEL_PROFILE_ALL));

	printf("Configured to send Grapheur messages to %s:%d\n", ADDRESS, PORT);
	printf("Multipliers (x, y, z): %f, %f, %f\n", mult_x, mult_y, mult_z);
	printf("Offsets (x, y, z): %f, %f, %f\n", off_x, off_y, off_z);

	printf("Initialized Kinect, looking for users...\n");

	context.StartGeneratingAll();

	// Initialize Winsock
	WSAStartup(MAKEWORD(2,2), &wsaData);
	SendSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	RecvAddr.sin_family = AF_INET;
	RecvAddr.sin_port = htons(PORT);
	RecvAddr.sin_addr.s_addr = inet_addr("127.0.0.1");

	broadcastScratch("broadcast kinect_server");			

	while (true) 
	{
		// Read next available data
		context.WaitAnyUpdateAll();
		// Process the data
		xn::DepthMetaData depthMD;
		depth.GetMetaData(depthMD);
		send(depthMD);
	}

	context.Shutdown();
	closesocket(SendSocket);
	WSACleanup();
	return 0;
}

